@@ -32,7 +32,7 @@ ensure_bitarray(PyObject *obj)
     return 0;
 }
 
-/* return new bitarray of length 'nbits' (with uninitialized buffer) and
+/* return new bitarray of length `nbits` (with uninitialized buffer) and
    endianness given by the PyObject 'endian' */
 static bitarrayobject *
 new_bitarray(Py_ssize_t nbits, PyObject *endian)
@@ -639,8 +639,8 @@ hex2ba_core(bitarrayobject *a, PyObject *bytes)
     return 0;
 }
 
-/* Return new reference to bytes object from either str (unicode) or bytes.
-   On failure, set exception and return NULL. */
+/* Return new reference to bytes object from either ASCII str (unicode) or
+   bytes.  On failure, set exception and return NULL. */
 static PyObject *
 anystr_to_bytes(PyObject *obj)
 {
@@ -669,14 +669,14 @@ hex2ba(PyObject *module, PyObject *args, PyObject *kwds)
 {
     static char *kwlist[] = {"", "endian", NULL};
     PyObject *obj, *bytes, *endian = Py_None;
-    bitarrayobject *a = NULL;
+    bitarrayobject *a;
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|O:hex2ba", kwlist,
                                      &obj, &endian))
         return NULL;
 
     if ((bytes = anystr_to_bytes(obj)) == NULL)
-        goto error;
+        return NULL;
 
     a = new_bitarray(4 * PyBytes_GET_SIZE(bytes), endian);
     if (a == NULL)
@@ -688,7 +688,7 @@ hex2ba(PyObject *module, PyObject *args, PyObject *kwds)
     Py_DECREF(bytes);
     return (PyObject *) a;
  error:
-    Py_XDECREF(bytes);
+    Py_DECREF(bytes);
     Py_XDECREF((PyObject *) a);
     return NULL;
 }
@@ -854,7 +854,7 @@ base2ba(PyObject *module, PyObject *args, PyObject *kwds)
 {
     static char *kwlist[] = {"", "", "endian", NULL};
     PyObject *obj, *bytes, *endian = Py_None;
-    bitarrayobject *a = NULL;
+    bitarrayobject *a;
     int n, m;                   /* n = 2^m */
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "iO|O:base2ba", kwlist,
@@ -865,7 +865,7 @@ base2ba(PyObject *module, PyObject *args, PyObject *kwds)
         return NULL;
 
     if ((bytes = anystr_to_bytes(obj)) == NULL)
-        goto error;
+        return NULL;
 
     a = new_bitarray(m * PyBytes_GET_SIZE(bytes), endian);
     if (a == NULL)
@@ -877,7 +877,7 @@ base2ba(PyObject *module, PyObject *args, PyObject *kwds)
     Py_DECREF(bytes);
     return (PyObject *) a;
  error:
-    Py_XDECREF(bytes);
+    Py_DECREF(bytes);
     Py_XDECREF((PyObject *) a);
     return NULL;
 }
@@ -1030,28 +1030,9 @@ byte_length(Py_ssize_t i)
     return n;
 }
 
-/* ---------------------- sparse compressed bitarray ------------------- */
-
-/* Bitarray buffer size (in bytes) that can be indexed by n bytes.  E.g.:
-   with 1 byte you can index 256 bits which have a buffer size of 32 bytes,
-   so BSI(1) = 32, BSI(2) = 8192, ... */
-#define BSI(n)  (((Py_ssize_t) 1) << (8 * (n) - 3))
-
-static int
-sc_encode_header(char *str, bitarrayobject *a)
-{
-    int len;
-
-    len = byte_length(a->nbits);
-    *str = (IS_BE(a) ? 0x10 : 0x00) | ((char) len);
-    write_n(str + 1, len, a->nbits);
-
-    return 1 + len;
-}
-
-/* starting at byte index 'i' count the remaining bits */
+/* starting from byte i count the remaining population in bitarray buffer */
 static Py_ssize_t
-count_final(bitarrayobject *a, Py_ssize_t i)
+count_from(bitarrayobject *a, Py_ssize_t i)
 {
     Py_ssize_t cnt = 0;
 
@@ -1064,12 +1045,21 @@ count_final(bitarrayobject *a, Py_ssize_t i)
     return cnt;
 }
 
-/* segment size in bytes - although of little practical value, the code
+/* ---------------------- sparse compressed bitarray -------------------
+ *
+ * see also: doc/sparse_compression.rst
+ */
+
+/* Bitarray buffer size (in bytes) that can be indexed by n bytes.  E.g.:
+   with 1 byte you can index 256 bits which have a buffer size of 32 bytes.
+   BSI(1) = 32, BSI(2) = 8_192, BSI(3) = 2_097_152, BSI(4) = 536_870_912 */
+#define BSI(n)  (((Py_ssize_t) 1) << (8 * (n) - 3))
+
+/* segment size in bytes - Although of little practical value, the code
    below will also work when changing SEGSIZE to 1, 2, 4, 8 or 16, as long
    as a multiple of SEGSIZE is 32.  The size 32 is rooted in the fact that
    a bitarray of 32 bytes (256 bits) can be indexed with one index byte
-   (BSI(0) = 32).  Our entire 'sc' format is constructed around this.
- */
+   (BSI(1) = 32).  Our entire 'sc' format is constructed around this. */
 #define SEGSIZE  32
 
 /* number of 256 bit segments given nbits */
@@ -1079,44 +1069,44 @@ count_final(bitarrayobject *a, Py_ssize_t i)
    Note that we call these "segments", as opposed to "blocks", in order to
    avoid confusion with encode blocks.
 
-   0       1       2       3       4       5    6   index into rts array, i
-   +-------+-------+-------+-------+-------+----+
-   |     5 |     7 |     0 |    11 |     6 |  4 |   count per segment
-   +-------+-------+-------+-------+-------+----+
-   0       5      12      12      23      29   33   running totals, rts[i]
+   0           1           2           3           4   index in rts array, i
+   +-----------+-----------+-----------+-----------+
+   |      5    |      0    |      3    |      4    |   segment population
+   |           |           |           |           |
+   |  [0:256]  | [256:512] | [512:768] | [768:987] |   bitarray slice
+   +-----------+-----------+-----------+-----------+
+   0           5           5           8          12   running totals, rts[i]
 
-   In this example we have a bitarray with 6 segments, e.g. with 1407 bits.
-   Note that:
+   In this example we have a bitarray of length nbits = 987.  Note that:
 
-     * Here we have 6 segments, but the rts array has a size
-       of 7 elements: 0 .. 6
-       The rts array has always NSEG(nbits) + 1 elements, such that the
-       last element is always indexed by NSEG(nbits).  Here, NSEG(1407) = 6
+     * The number of segments is given by NSEG(nbits).
+       Here we have 4 segments: NSEG(nbits) = NSEG(987) = 4
 
-     * The zeroth element rts[0] is always zero.
+     * The rts array has always NSEG(nbits) + 1 elements, such that
+       last element is always indexed by NSEG(nbits).
 
-     * The last last element rts[NSEG(nbits)] is always the total count.
-       Here: rts[NSEG(nbits)] = rts[NSEG(1407)] = rts[6] = 33
+     * The element rts[0] is always zero.
 
-     * The last segment may be partial.  Here, spanning 127 bits, that
-       is a[1280:1407].  The count of this segment is 4:
-       a[1280:].count() = a.count(1, 1280) = 4
+     * The last element rts[NSEG(nbits)] is always the total count.
+       Here: rts[NSEG(nbits)] = rts[NSEG(987)] = rts[4] = 12
 
-     * The segment a[512:768] has a count of zero, such that: rts[2] = rts[3]
+     * The last segment may be partial.  In that case, it's size it given
+       by nbits % 256.  Here: nbits % 256 = 987 % 256 = 219
 
-   As each segment covers 256 bits (32 bytes), and each element in the
-   running totals array takes up 8 bytes (on a 64-bit machine at least) the
-   additional memory to accommodate the rts array is 1/4 of the bitarrays
-   memory.  However, calculating this array upfront allows count_block() to
+   As each segment (at large) covers 256 bits (32 bytes), and each element
+   in the running totals array takes up 8 bytes (on a 64-bit machine) the
+   additional memory to accommodate the rts array is therefore 1/4 of the
+   bitarray's memory.
+   However, calculating this array upfront allows sc_count() to
    simply look up two entries from the array and take their difference.
-   Thus, the speedup we get can be significant.
+   Thus, the speedup is significant.
 
-   The function write_sparse_block() also takes advantage of the running
-   totals array, as it can quickly check for segments that only contain
-   only zeros to skip.
+   The function sc_write_indices() also takes advantage of the running
+   totals array.  It loops over segments and skips to the next segment as
+   soon as the count the current segment is reached.
 */
 static Py_ssize_t *
-calc_rts(bitarrayobject *a)
+sc_calc_rts(bitarrayobject *a)
 {
     const Py_ssize_t n_seg = NSEG(a->nbits);  /* number of segments */
     const Py_ssize_t c_seg = a->nbits / (8 * SEGSIZE);  /* complete segs */
@@ -1125,6 +1115,7 @@ calc_rts(bitarrayobject *a)
     char *buff;
     Py_ssize_t *res, i, j;
 
+    assert(n_seg == c_seg || n_seg == c_seg + 1);
     memset(zeros, 0x00, SEGSIZE);
     res = (Py_ssize_t *) PyMem_Malloc((size_t)
                                       sizeof(Py_ssize_t) * (n_seg + 1));
@@ -1144,11 +1135,10 @@ calc_rts(bitarrayobject *a)
     res[c_seg] = cnt;
 
     if (n_seg > c_seg) {           /* we have a final partial segment */
-        assert(n_seg == c_seg + 1);
         assert(Py_SIZE(a) <= SEGSIZE * n_seg);
         assert(a->nbits && a->nbits < 8 * SEGSIZE * n_seg);
 
-        cnt += count_final(a, SEGSIZE * c_seg);
+        cnt += count_from(a, SEGSIZE * c_seg);
         res[n_seg] = cnt;
     }
     return res;
@@ -1161,7 +1151,7 @@ calc_rts(bitarrayobject *a)
    The offset is required to be multiple of 32 (the segment size), as this
    functions makes use of running totals (stored in Py_ssize_t array rts). */
 static Py_ssize_t
-count_block(bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset, int n)
+sc_count(bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset, int n)
 {
     Py_ssize_t nbits;
 
@@ -1178,8 +1168,7 @@ count_block(bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset, int n)
        is 1 << 32.  This is problematic, as 32-bit machines can address
        at least partially filled type 4 blocks).  Therefore, we first
        limit BSI(n) by the buffer size before multiplying 8. */
-    nbits = Py_MIN(8 * Py_MIN(BSI(n), Py_SIZE(a)),
-                   a->nbits - 8 * offset);
+    nbits = Py_MIN(8 * Py_MIN(BSI(n), Py_SIZE(a)), a->nbits - 8 * offset);
     assert(nbits >= 0);
 
     offset /= SEGSIZE;               /* offset in terms of segments now */
@@ -1193,10 +1182,9 @@ count_block(bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset, int n)
    buffer str), and return the number of raw bytes.
    Note that the encoded block size is the return value + 1. */
 static int
-write_raw_block(char *str, bitarrayobject *a, Py_ssize_t *rts,
-                Py_ssize_t offset)
+sc_write_raw(char *str, bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset)
 {
-    Py_ssize_t nbytes = Py_SIZE(a) - offset;  /* remaining bytes */
+    const Py_ssize_t nbytes = Py_SIZE(a) - offset;  /* remaining bytes */
     Py_ssize_t k = Py_MIN(32, nbytes);
 
     assert(nbytes > 0);
@@ -1205,8 +1193,10 @@ write_raw_block(char *str, bitarrayobject *a, Py_ssize_t *rts,
            raw bytes (otherwise this function wouldn't have been called).
            Now also check the next 3 segments. */
         while (k < 128 &&
-               Py_MIN(32, nbytes - k) <= count_block(a, rts, offset + k, 1))
+               Py_MIN(32, nbytes - k) <= sc_count(a, rts, offset + k, 1))
             k += 32;
+
+        assert(k % SEGSIZE == 0);
     }
     k = Py_MIN(k, nbytes);
     assert(0 < k && k <= 128 && k <= nbytes);
@@ -1221,56 +1211,87 @@ write_raw_block(char *str, bitarrayobject *a, Py_ssize_t *rts,
     return (int) k;
 }
 
-/* Encode one sparse block (from offset, and up to k one bits).
-   Return number of bytes written to buffer str (encoded block size). */
-static Py_ssize_t
-write_sparse_block(char *str, bitarrayobject *a, Py_ssize_t *rts,
-                   Py_ssize_t offset, int n, int k)
+/* Write `k` indices (of `n` bytes each) into buffer `str`.
+   Note that `n` (which is also the block type) has been selected
+   (in sc_encode_block()) such that:
+
+       k = sc_count(a, rts, offset, n) < 256
+*/
+static void
+sc_write_indices(char *str, bitarrayobject *a, Py_ssize_t *rts,
+                 Py_ssize_t offset, int n, int k)
 {
-    /* bytes to encode limited by remaining buffer size */
-    Py_ssize_t na = Py_MIN(BSI(n), Py_SIZE(a) - offset);
-    Py_ssize_t i, j, outsize, len = 0;
-    char *buff = a->ob_item + offset;
+    const char *str_stop = str + n * k;  /* stop position in buffer `str` */
+    const char *buff = a->ob_item + offset;
+    Py_ssize_t m, i;
 
+    assert(1 <= n && n <= 4);
+    assert(0 < k && k < 256);  /* note that k cannot be 0 in this function */
+    assert(k == sc_count(a, rts, offset, n));   /* see above */
     assert(offset % SEGSIZE == 0);
-    assert(1 <= n && n <= 4 && 0 <= k && k < 256);
-    assert(na > 0 && offset + na <= Py_SIZE(a));
 
-    /* block header */
+    rts += offset / SEGSIZE;   /* rts index relative to offset now */
+
+    for (m = 0;;) {  /* loop segments */
+        int j, ni;
+
+        assert(m + offset / SEGSIZE < NSEG(a->nbits));
+        ni = (int) (rts[m + 1] - rts[m]);  /* indices in this segment */
+        if (ni == 0)
+            goto next_segment;
+
+        for (i = m * SEGSIZE;; i++) {  /* loop bytes in segment */
+            assert(i < (m + 1) * SEGSIZE && i + offset < Py_SIZE(a));
+            if (buff[i] == 0)
+                continue;
+
+            for (j = 0; j < 8; j++)  /* loop bits */
+                if (buff[i] & BITMASK(a, j)) {
+                    write_n(str, n, 8 * i + j);
+                    str += n;
+                    if (--ni == 0) {
+                        /* we have encountered all indices in this segment */
+                        if (str == str_stop)
+                            return;
+                        goto next_segment;
+                    }
+                }
+        }
+    next_segment:
+        m++;
+    }
+    Py_UNREACHABLE();
+}
+
+#undef SEGSIZE
+#undef NSEG
+
+/* Write one sparse block (from `offset`, and up to `k` one bits).
+   Return number of bytes written to buffer `str` (encoded block size). */
+static Py_ssize_t
+sc_write_sparse(char *str, bitarrayobject *a, Py_ssize_t *rts,
+                Py_ssize_t offset, int n, int k)
+{
+    int len = 0;
+
+    assert(1 <= n && n <= 4);
+    assert(0 <= k && k < 256);
+
+    /* write block header */
     if (n == 1) {               /* type 1 - single byte for each position */
         assert(k < 32);
-        str[len++] = 0xa0 + k;
+        str[len++] = (char) (0xa0 + k);
     }
-    else {            /* type 2, 3, 4 - multiple bytes for each positions */
-        str[len++] = 0xc0 + n;
-        str[len++] = k;
+    else {                 /* type 2, 3, 4 - `n` bytes for each positions */
+        str[len++] = (char) (0xc0 + n);
+        str[len++] = (char) k;
     }
     if (k == 0)  /* no index bytes */
         return len;
 
-    /* block data */
-    outsize = len + n * k;
-    for (i = 0; i < na; i++) {
-        if (i % SEGSIZE == 0) {
-            j = (i + offset) / SEGSIZE;  /* running total index */
-            assert(j < NSEG(a->nbits));
-            if (rts[j] == rts[j + 1]) {  /* the segment has only zeros */
-                i += SEGSIZE - 1;        /* skip ahead */
-                continue;
-            }
-        }
-        assert(offset + i < Py_SIZE(a));
-        if (buff[i])
-            for (j = 0; j < 8; j++)
-                if (buff[i] & BITMASK(a, j)) {
-                    write_n(str + len, n, 8 * i + j);
-                    len += n;
-                    if (len == outsize)  /* final index reached */
-                        return len;
-                }
-    }
-    Py_FatalError("internal sc_encode() error");
-    return -1;  /* cannot happen - silence compiler warning */
+    /* write block data - `k` indices, `n` bytes per index */
+    sc_write_indices(str + len, a, rts, offset, n, k);
+    return len + n * k;
 }
 
 /* Encode one block (starting at offset) and return offset increment.
@@ -1312,15 +1333,15 @@ static Py_ssize_t
 sc_encode_block(char *str, Py_ssize_t *len,
                 bitarrayobject *a, Py_ssize_t *rts, Py_ssize_t offset)
 {
-    Py_ssize_t nbytes = Py_SIZE(a) - offset;        /* remaining bytes */
+    const Py_ssize_t nbytes = Py_SIZE(a) - offset;  /* remaining bytes */
     int count, n;
 
     assert(nbytes > 0);
 
-    count = (int) count_block(a, rts, offset, 1);
+    count = (int) sc_count(a, rts, offset, 1);
     /* are there fewer or equal raw bytes than index bytes */
     if (Py_MIN(32, nbytes) <= count) {           /* type 0 - raw bytes */
-        int k = write_raw_block(str + *len, a, rts, offset);
+        int k = sc_write_raw(str + *len, a, rts, offset);
         *len += 1 + k;
         return k;
     }
@@ -1329,7 +1350,7 @@ sc_encode_block(char *str, Py_ssize_t *len,
         Py_ssize_t next_count, size_a, size_b;
 
         /* population for next block type n+1 */
-        next_count = count_block(a, rts, offset, n + 1);
+        next_count = sc_count(a, rts, offset, n + 1);
         if (next_count >= 256)
             /* too many index bytes for next block type n+1 */
             break;
@@ -1347,12 +1368,21 @@ sc_encode_block(char *str, Py_ssize_t *len,
         count = (int) next_count;
     }
 
-    *len += write_sparse_block(str + *len, a, rts, offset, n, count);
+    *len += sc_write_sparse(str + *len, a, rts, offset, n, count);
     return BSI(n);
 }
 
-#undef SEGSIZE
-#undef NSEG
+static int
+sc_encode_header(char *str, bitarrayobject *a)
+{
+    int len;
+
+    len = byte_length(a->nbits);
+    *str = (IS_BE(a) ? 0x10 : 0x00) | ((char) len);
+    write_n(str + 1, len, a->nbits);
+
+    return 1 + len;
+}
 
 static PyObject *
 sc_encode(PyObject *module, PyObject *obj)
@@ -1361,14 +1391,15 @@ sc_encode(PyObject *module, PyObject *obj)
     char *str;                  /* output buffer */
     Py_ssize_t len = 0;         /* bytes written into output buffer */
     bitarrayobject *a;
-    Py_ssize_t offset = 0;      /* block offset into bitarray a in bytes */
+    Py_ssize_t offset = 0;      /* block offset into bitarray `a` in bytes */
     Py_ssize_t *rts;            /* running totals for 256 bit segments */
 
     if (ensure_bitarray(obj) < 0)
         return NULL;
 
     a = (bitarrayobject *) obj;
-    if ((rts = calc_rts(a)) == NULL)
+    set_padbits(a);
+    if ((rts = sc_calc_rts(a)) == NULL)
         return NULL;
 
     out = PyBytes_FromStringAndSize(NULL, 32768);
@@ -1383,7 +1414,7 @@ sc_encode(PyObject *module, PyObject *obj)
 
         /* Make sure we have enough space in output buffer for next block.
            The largest block possible is a type 4 block with 255 indices.
-           It's site is: 2 header bytes + 4 * 255 index bytes */
+           It's size is: 2 header bytes + 4 * 255 index bytes */
         allocated = PyBytes_GET_SIZE(out);
         if (allocated < len + 2 + 4 * 255) {  /* increase allocation */
             if (_PyBytes_Resize(&out, allocated + 32768) < 0)
@@ -1440,10 +1471,10 @@ sc_decode_header(PyObject *iter, int *endian, Py_ssize_t *nbits)
 /* Read k bytes from iter and set elements in bitarray.  Return the size of
    the offset increment in bytes (i.e. just k), or -1 on failure. */
 static Py_ssize_t
-read_raw_block(bitarrayobject *a, Py_ssize_t offset, PyObject *iter, int k)
+sc_read_raw(bitarrayobject *a, Py_ssize_t offset, PyObject *iter, int k)
 {
-    Py_ssize_t i;
     char *buff = a->ob_item + offset;
+    Py_ssize_t i;
 
     if (offset + k > Py_SIZE(a)) {
         PyErr_Format(PyExc_ValueError, "decode error (raw): %zd + %d > %zd",
@@ -1463,8 +1494,8 @@ read_raw_block(bitarrayobject *a, Py_ssize_t offset, PyObject *iter, int k)
 /* Read n * k bytes from iter and set elements in bitarray.
    Return size of offset increment in bytes, or -1 on failure. */
 static Py_ssize_t
-read_sparse_block(bitarrayobject *a, Py_ssize_t offset, PyObject *iter,
-                  int n, int k)
+sc_read_sparse(bitarrayobject *a, Py_ssize_t offset, PyObject *iter,
+               int n, int k)
 {
     assert(1 <= n && n <= 4 && k >= 0);
     while (k--) {
@@ -1499,10 +1530,10 @@ sc_decode_block(bitarrayobject *a, Py_ssize_t offset, PyObject *iter)
         return 0;
 
     if (head <= 0x80)                      /* type 0 - 0x01 .. 0x80 */
-        return read_raw_block(a, offset, iter, head);
+        return sc_read_raw(a, offset, iter, head);
 
     if (0xa0 <= head && head < 0xc0)       /* type 1 - 0xa0 .. 0xbf */
-        return read_sparse_block(a, offset, iter, 1, head - 0xa0);
+        return sc_read_sparse(a, offset, iter, 1, head - 0xa0);
 
     if (0xc2 <= head && head <= 0xc4) {    /* type 2 .. 4 - 0xc2 .. 0xc4 */
         int k;
@@ -1510,7 +1541,7 @@ sc_decode_block(bitarrayobject *a, Py_ssize_t offset, PyObject *iter)
         if ((k = next_char(iter)) < 0)
             return -1;
 
-        return read_sparse_block(a, offset, iter, head - 0xc0, k);
+        return sc_read_sparse(a, offset, iter, head - 0xc0, k);
     }
 
     PyErr_Format(PyExc_ValueError, "invalid block head: 0x%02x", head);
