@@ -948,21 +948,22 @@ class TestsHexlify(unittest.TestCase, Util):
             self.assertEqual(ba2hex(a_be), hex_be)
             self.assertEqual(ba2hex(a_le), hex_le)
 
-    def test_round_trip(self):
-        s = ''.join(choice(hexdigits) for _ in range(randint(20, 100)))
+    def test_hexdigits(self):
         for default_endian in 'big', 'little':
             _set_default_endian(default_endian)
-            a = hex2ba(s)
-            self.check_obj(a)
+            a = hex2ba(hexdigits)
             self.assertEqual(len(a) % 4, 0)
             self.assertEqual(a.endian(), default_endian)
+            self.assertIsType(a, 'bitarray')
+            self.check_obj(a)
+
             t = ba2hex(a)
-            self.assertEqual(t, s.lower())
-            b = hex2ba(t, default_endian)
-            self.assertEQUAL(a, b)
+            self.assertEqual(t, hexdigits.lower())
+            self.assertIsInstance(t, str)
+            self.assertEQUAL(a, hex2ba(t, default_endian))
 
     def test_binascii(self):
-        a = urandom(800, 'big')
+        a = urandom(80, 'big')
         s = binascii.hexlify(a.tobytes()).decode()
         self.assertEqual(ba2hex(a), s)
         b = bitarray(endian='big')
@@ -1145,8 +1146,7 @@ class SC_Tests(unittest.TestCase, Util):
                 (b'\x11\x09\xa1\x08\0',     '00000000 1',        'big'),
                 (b'\x01E\xa3ABD\0',         65 * '0' + '1101',   'little'),
         ]:
-            # the padbits in a frozenbitarray are guaranteed to be zero
-            a = frozenbitarray(bits, endian)
+            a = bitarray(bits, endian)
             self.assertEqual(sc_encode(a), b)
             self.assertEqual(sc_decode(b), a)
 
@@ -1345,16 +1345,16 @@ class SC_Tests(unittest.TestCase, Util):
         for a in bitarray('1', 'big'), frozenbitarray('1', 'big'):
             b = sc_encode(a)
             self.assertIsInstance(b, bytes)
-            self.assertEqual(len(b), 5)
+            self.assertEqual(b, b'\x11\x01\x01\x80\0')
 
         for a in None, [], 0, 123, b'', b'\x00', 3.14:
             self.assertRaises(TypeError, sc_encode, a)
 
     def round_trip(self, a):
-        b = sc_encode(a)
-        c = sc_decode(b)
-        self.assertEqual(a, c)
-        self.assertEqual(a.endian(), c.endian())
+        c = a.copy()
+        b = sc_decode(sc_encode(a))
+        self.assertTrue(a == b == c)
+        self.assertTrue(a.endian() == b.endian() == c.endian())
 
     def test_encode_zeros(self):
         for i in range(18):
@@ -1523,10 +1523,10 @@ class VLFTests(unittest.TestCase, Util):
             self.assertEqual(vl_decode(s), a)
 
     def round_trip(self, a):
+        c = a.copy()
         s = vl_encode(a)
         b = vl_decode(s)
-        self.check_obj(b)
-        self.assertEqual(a, b)
+        self.assertTrue(a == b == c)
         LEN_PAD_BITS = 3
         self.assertEqual(len(s), (len(a) + LEN_PAD_BITS + 6) // 7)
 
